//
//  CreepyGirlController.m
//  CreepyGirl
//
//  Created by Curtis Jones on 2008.03.31.
//  Copyright 2008 __MyCompanyName__. All rights reserved.
//

#import "CreepyGirlController.h"

@implementation CreepyGirlController

- (id)init
{
	self = [super init];
	
	if (self) {
		mFrameCount = 0;
		mMotionTracker = [[MotionTracker alloc] init];
	}
	
	return self;
}

- (void)dealloc
{
	[mMotionTracker release];
	mMotionTracker = nil;
	
	[mCaptureSession release];
	mCaptureSession = nil;
	
	[super dealloc];
}

/**
 *
 *
 */
- (void)awakeFromNib
{
	NSError * error;
	QTCaptureDecompressedVideoOutput *videoOutput;
	
	[NSApp setDelegate:self];
	
	// create the capture session
	mCaptureSession = [[QTCaptureSession alloc] init];
	
	// add a decompressed video output that returns raw frames to the session
	videoOutput = [[QTCaptureDecompressedVideoOutput alloc] init];
	[videoOutput setDelegate:self];
	
	if (![mCaptureSession addOutput:videoOutput error:&error]) {
		NSLog(@"%s.. failed to [mCaptureSession addOutput:videoOutput error:&error], %d", __PRETTY_FUNCTION__, [error code]);
		return;
	}
	
	[mCaptureView setCaptureSession:mCaptureSession];
	
	NSLog(@"%s.. ready!", __PRETTY_FUNCTION__);
}

- (void)applicationWillTerminate:(NSNotification *)aNotification
{
	NSLog(@"%s..", __PRETTY_FUNCTION__);
	
	[self stopTracking:nil];
}

/**
 * Test moving the mouse.
 *
 */
- (IBAction)moveMouse:(id)sender
{
	NSLog(@"%s.. ", __PRETTY_FUNCTION__);
	
	NSPoint point = {100.0, 50.0};
	NSRect  rect  = {{0.0, 0.0},{640.0, 480.0}};
	
	[self positionMouseAtPoint:point forView:mWebView withRect:rect];
}

/**
 *
 *
 */
- (IBAction)testScan:(id)sender
{
	NSLog(@"%s..", __PRETTY_FUNCTION__);
	
	[mMotionTracker testScan001];
}

/**
 *
 *
 */
- (IBAction)testCompare:(id)sender
{
	NSLog(@"%s..", __PRETTY_FUNCTION__);
	
	[mMotionTracker testCompare001];
}

/**
 *
 *
 */
- (IBAction)testLoad:(id)sender
{
	NSLog(@"%s..", __PRETTY_FUNCTION__);
	
//[[mWebView mainFrame] loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:@"file:///Users/cjones/Projects/Personal/CreepyGirl/TIminoriHair.swf"]]];
	[[mWebView mainFrame] loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:@"http://www.motionportrait.com/about/TIminoriHair.swf"]]];
}

/**
 *
 *
 */
- (IBAction)startTracking:(id)sender
{
	NSLog(@"%s..", __PRETTY_FUNCTION__);
	
	NSError * error;
	
	[mBrowserWin makeKeyAndOrderFront:nil];
	[mBrowserWin makeFirstResponder:nil];
	
	// find a video device
	mVideoDevice = [QTCaptureDevice defaultInputDeviceWithMediaType:QTMediaTypeVideo];
	
	if (mVideoDevice == nil || ![mVideoDevice open:&error]) {
		NSLog(@"%s.. failed to [QTCaptureDevice defaultInputDeviceWithMediaType:QTMediaTypeVideo], %d", __PRETTY_FUNCTION__, [error code]);
		
		mVideoDevice = [QTCaptureDevice defaultInputDeviceWithMediaType:QTMediaTypeMuxed];
		
		if (mVideoDevice == nil || ![mVideoDevice open:&error]) {
			NSLog(@"%s.. failed to [QTCaptureDevice defaultInputDeviceWithMediaType:QTMediaTypeMuxed], %d", __PRETTY_FUNCTION__, [error code]);
			return;
		}
	}
	
	// add the video device to the session as a device input.
	mVideoInput = [[QTCaptureDeviceInput alloc] initWithDevice:mVideoDevice];
	
	if (![mCaptureSession addInput:mVideoInput error:&error]) {
		NSLog(@"%s.. failed to [[QTCaptureDeviceInput alloc] initWithDevice:device], %d", __PRETTY_FUNCTION__, [error code]);
		return;
	}
	
	// if the video device doesn't also supply audio, add an audio device input 
	// to the session.
	if (![mVideoDevice hasMediaType:QTMediaTypeSound] && ![mVideoDevice hasMediaType:QTMediaTypeMuxed]) {
		QTCaptureDevice * audioDevice = [QTCaptureDevice defaultInputDeviceWithMediaType:QTMediaTypeSound];
		
		if (audioDevice == nil || ![audioDevice open:&error]) {
			NSLog(@"%s.. failed to [QTCaptureDevice defaultInputDeviceWithMediaType:QTMediaTypeSound], %d", __PRETTY_FUNCTION__, [error code]);
			return;
		}
		
		mAudioInput = [[QTCaptureDeviceInput alloc] initWithDevice:audioDevice];
		
		if (![mCaptureSession addInput:mAudioInput error:&error]) {
			NSLog(@"%s.. failed to [[QTCaptureDeviceInput alloc] initWithDevice:audio], %d", __PRETTY_FUNCTION__, [error code]);
			return;
		}
	}
	
	[mCaptureSession startRunning];
}

/**
 *
 *
 */
- (IBAction)stopTracking:(id)sender
{
	NSLog(@"%s..", __PRETTY_FUNCTION__);
	
	[mCaptureSession stopRunning];
	
	if ([[mVideoInput device] isOpen])
		[[mVideoInput device] close];
	
	if ([[mAudioInput device] isOpen])
		[[mAudioInput device] close];
}

/**
 * this delegate method is called whenever the QTCaptureDecompressedVideoOutput 
 * receives a frame.
 *
 */
- (void)captureOutput:(QTCaptureOutput *)captureOutput didOutputVideoFrame:(CVImageBufferRef)videoFrame withSampleBuffer:(QTSampleBuffer *)sampleBuffer fromConnection:(QTCaptureConnection *)connection
{
	NSPoint point;
	
	mFrameCount++;
	
	@synchronized (self)
	{
		CIImage * ciimage = [CIImage imageWithCVImageBuffer:videoFrame];
		NSBitmapImageRep * bitmap = [self RGBABitmapImageRepWithCImage:ciimage];
		
		int size = (([bitmap pixelsHigh] * [bitmap pixelsWide]) * [bitmap bitsPerPixel]) >> 3; size;
//	NSLog(@"%s.. bytes=%d", __PRETTY_FUNCTION__, size);
		
//	point = [mMotionTracker addFrameData:[bitmap bitmapData] width:[bitmap pixelsWide] height:[bitmap pixelsHigh] bitsPerPixel:[bitmap bitsPerPixel]];
//	[self positionMouseAtPoint:point forView:mWebView withRect:(NSRect){{0.0,0.0}, {(float)[bitmap pixelsWide],(float)[bitmap pixelsHigh]}}];
		
//	NSLog(@"%s.. pixelsWide=%d, pixelsHigh=%d", __PRETTY_FUNCTION__, [bitmap pixelsWide], [bitmap pixelsHigh]);
//	NSLog(@"%s.. point.x=%d, point.y=%d", __PRETTY_FUNCTION__, (int)point.x, (int)point.y);
	}
	
//if (mFrameCount >= 50)
//	[self stopTracking:nil];
}

/**
 * CGGLContextCreate
 *
 */
- (NSBitmapImageRep *)RGBABitmapImageRepWithCImage:(CIImage *) ciImage
{
	int width = [ciImage extent].size.width;
	int rows = [ciImage extent].size.height;
	int rowBytes = (width * 4);
	
	NSBitmapImageRep* rep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:nil pixelsWide:width pixelsHigh:rows bitsPerSample:8 samplesPerPixel:4 hasAlpha:YES isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace bitmapFormat:0 bytesPerRow:rowBytes bitsPerPixel:0];
	
	CGColorSpaceRef colorSpace = CGColorSpaceCreateWithName ( kCGColorSpaceGenericRGB );
	CGContextRef context = CGBitmapContextCreate( [rep bitmapData], width, rows, 8, rowBytes, colorSpace, kCGImageAlphaPremultipliedLast );
	
	CIContext* ciContext = [CIContext contextWithCGContext:context options:nil];
	[ciContext drawImage:ciImage atPoint:CGPointZero fromRect: [ciImage extent]];
	
 	CGContextRelease( context );
	CGColorSpaceRelease( colorSpace );
	
	return [rep autorelease];
}

/**
 * Positions the mouse in the specified location based on the top-left corner
 * being the 0,0 position.
 *
 * Cocoa is all based around the bottom-left corner being 0,0, so this 
 * difference is important to remember.
 *
 */
- (void)positionMouseAtScreenPoint:(NSPoint)screenPoint
{
	CGPoint point;
	CGError error;
	
	point.x = screenPoint.x;
	point.y = screenPoint.y;
	
	if (0 != (error = CGPostMouseEvent(point, TRUE, 1, FALSE)))
		NSLog(@"%s.. error=%d", __PRETTY_FUNCTION__, error);
}

/**
 * Position the mouse at aPoint in aView based on first transforming the point
 * based on resizing aRect to match the rect of aView and then translating the 
 * point to the screen coordinate system.
 *
 */
- (void)positionMouseAtPoint:(NSPoint)aPoint forView:(NSView *)aView withRect:(NSRect)aRect
{
	NSRect viewRect = [aView frame];
	
	// translate the point relative to the size difference between aRect and the
	// destination rect of aView.
	//
	if (aPoint.x != 0.0)
		aPoint.x = (aPoint.x * viewRect.size.width ) / aRect.size.width;
	
	if (aPoint.y != 0.0);
		aPoint.y = (aPoint.y * viewRect.size.height) / aRect.size.height;
	
	// translate the point from the coordinate system of the view to the 
	// coordinate system of the parent window and then to the coordinate system
	// of the screen itself.
	//
	aPoint = [aView convertPointToBase:aPoint];
	aPoint = [[aView window] convertBaseToScreen:aPoint];
	
	// translate from a bottom-left origin to a top-left origin.
	//
	aPoint.y = [[[aView window] screen] frame].size.height - aPoint.y;
	
//NSLog(@"%s.. point.x=%d, point.y=%d", __PRETTY_FUNCTION__, (int)aPoint.x, (int)aPoint.y);
	
	// move the mouse to the desired point, finally.
	//
	[self positionMouseAtScreenPoint:aPoint];
}

@end
