//
//  MotionTracker.m
//  CreepyGirl
//
//  Created by Curtis Jones on 2008.03.31.
//  Copyright 2008 __MyCompanyName__. All rights reserved.
//

#import "MotionTracker.h"
#import <errno.h>
#import <stdio.h>
#import <stdlib.h>
#import <string.h>

@implementation MotionTracker

/**
 *
 *
 */
- (id)init
{
	self = [super init];
	
	if (self) {
		mLastPoint.x = 0.0;
		mLastPoint.y = 0.0;
		mVertDir = 1;
		mHorzDir = 1;
		
		mFrameNew = (unsigned int *)malloc(sizeof(int) * IMAGE_COLS * IMAGE_ROWS);
		mFrameOld = (unsigned int *)malloc(sizeof(int) * IMAGE_COLS * IMAGE_ROWS);
		mMarks		= (CreepyPixel *)malloc(sizeof(CreepyPixel) * IMAGE_COLS * IMAGE_ROWS);
		
		memset(mFrameNew, 0, sizeof(int) * IMAGE_COLS * IMAGE_ROWS);
		memset(mFrameOld, 0, sizeof(int) * IMAGE_COLS * IMAGE_ROWS);
		memset(mMarks, 0, sizeof(CreepyPixel) * IMAGE_COLS * IMAGE_ROWS);
		memset(&mCurTrackedObject, 0, sizeof(TrackedObject));
	}
	
	return self;
}

/**
 *
 *
 */
- (void)dealloc
{
	if (mMarks != NULL) {
		free(mMarks);
		mMarks = NULL;
	}
	
	if (mFrameNew != NULL) {
		free(mFrameNew);
		mFrameNew = NULL;
	}
	
	if (mFrameOld != NULL) {
		free(mFrameOld);
		mFrameOld = NULL;
	}
	
	[super dealloc];
}

/**
 * Create marks and identify the objects.
 *
 */
- (void)testScan001
{
	NSLog(@"%s..", __PRETTY_FUNCTION__);
	
	int frameCount=0, rowStart=100, colStart=100, rowIncrement=3, colIncrement=3;
	
	for (; frameCount < 10; ++frameCount, rowStart+=rowIncrement, colStart+=colIncrement) {
		// mark a block of pixels as though that whole area had been seen to move 
		// from the old frame to the current frame.
		//
		{
			int row, col;
			
			memset(mMarks, 0, sizeof(CreepyPixel) * IMAGE_COLS * IMAGE_ROWS);
			
			for (row = rowStart; row < rowStart+150; ++row) {
				for (col = colStart; col < colStart+150; ++col) {
					mMarks[(IMAGE_COLS*row)+col].mark = 1;
				}
				for (col = colStart+200; col < colStart+250; ++col) {
					mMarks[(IMAGE_COLS*row)+col].mark = 1;
				}
			}
		}
		
		[self scanMarksForFrame:mMarks];
		[self calculateCentersAndAreasForObjects:mTrackedObjects arraySize:mTrackedObjectCount];
		
		{
			NSLog(@"%s.. found %d distinct objects.", __PRETTY_FUNCTION__, mTrackedObjectCount);
			
			int objCnt = 0;
			
			for (; objCnt < mTrackedObjectCount; ++objCnt) {
				TrackedObject * object = &mTrackedObjects[objCnt];
				NSLog(@"%s..   [%02d of %02d] high=%d, low=%d, left=%d, right=%d, area=%d, center.x=%d, center.y=%d", __PRETTY_FUNCTION__, 
							(objCnt+1), mTrackedObjectCount, object->high, object->low, object->left, object->right,
							object->area, (int)object->center.x, (int)object->center.y);
			}
		}
		
		// find where we think our currently tracked object has moved to,
		// if it has moved at all, or assign one of the objects to be the
		// currently tracked object if we don't already have one assigned.
		//
		if (mTrackedObjectCount == 0)
			NSLog(@"%s.. we couldn't find any movement. sorry.", __PRETTY_FUNCTION__);
		else if (mCurTrackedObject.center.x == 0.0 || mCurTrackedObject.center.y == 0.0)
			mCurTrackedObject = [self findMostProminentObject:(TrackedObject *)mTrackedObjects objectCount:mTrackedObjectCount];
		else
			mCurTrackedObject = [self findObjectLike:mCurTrackedObject fromObjects:mTrackedObjects objectCount:mTrackedObjectCount];
		
		NSLog(@"%s.. object->{area=%d, x=%d, y=%d}", __PRETTY_FUNCTION__, mCurTrackedObject.area, (int)mCurTrackedObject.center.x, (int)mCurTrackedObject.center.y);
	}
}

/**
 * Test comparing two frames that differ slightly from each other.
 *
 */
- (void)testCompare001
{
	NSLog(@"%s..", __PRETTY_FUNCTION__);
	
	// populate mFrameNew, mFrameOld with slightly shifted data
	// compare and place result into mMarks
	// analyze and validate mMarks
	
	NSPoint point;
	unsigned int *frameData;
	int frameCount=0, rowStart=100, colStart=100, rowIncrement=3, colIncrement=3;
	
	if (NULL == (frameData = malloc(IMAGE_COLS * IMAGE_ROWS * 4))) {
		NSLog(@"%s.. failed to malloc, %s", __PRETTY_FUNCTION__, strerror(errno));
		return;
	}
	
	for (; frameCount < 10; ++frameCount, rowStart+=rowIncrement, colStart+=colIncrement) {
		int row, col;
		
		memset(frameData, 0, sizeof(int) * IMAGE_COLS * IMAGE_ROWS);
		
		/*
		for (row = 0; row < IMAGE_ROWS; ++row) {
			for (col = 0; col < IMAGE_COLS; ++col) {
				frameData[(IMAGE_COLS*row)+col] = 0x123456FF;
			}
		}
		*/
		
		for (row = rowStart; row < rowStart+150; ++row) {
			for (col = colStart; col < colStart+150; ++col) {
				frameData[(IMAGE_COLS*row)+col] = 0x12345600;
			}
			for (col = colStart+200; col < colStart+250; ++col) {
				frameData[(IMAGE_COLS*row)+col] = 0x23456700;
			}
		}
		
		point = [self addFrameData:(unsigned char *)frameData width:IMAGE_COLS height:IMAGE_ROWS bitsPerPixel:4];
		
		NSLog(@"%s.. row=%d, col=%d", __PRETTY_FUNCTION__, (int)point.y, (int)point.x);
	}
}

/**
 * Find the object in the objects array that is most like the target object in
 * terms of both area and proximity.
 *
 * The result of this search will be our best guess on where the object in the
 * old frame moved to in the new frame.
 *
 */
- (TrackedObject)findObjectLike:(TrackedObject)targetObject fromObjects:(TrackedObject *)objectsArray objectCount:(int)objectCount
{
//NSLog(@"%s.. looking for object like targetObject->{area=%d, x=%d, y=%d}", __PRETTY_FUNCTION__, targetObject.area, (int)targetObject.center.x, (int)targetObject.center.y);
	
	int objectArrayPos=0, bestObjectScore=-1;
	TrackedObject *objectsArrayPtr = objectsArray;
	TrackedObject bestObject;
	
	memset(&bestObject, 0, sizeof(TrackedObject));
	
	for (; objectArrayPos < objectCount; ++objectArrayPos, ++objectsArrayPtr) {
		int areaScore = abs(targetObject.area - objectsArrayPtr[0].area);
		int locationScore = abs((targetObject.center.x - objectsArrayPtr[0].center.x) +
														(targetObject.center.y - objectsArrayPtr[0].center.y));
		int score = (areaScore * AREA_WEIGHT) + (locationScore * DISTANCE_WEIGHT);
		
		if (bestObjectScore == -1 || score < bestObjectScore) {
			bestObjectScore = score;
			bestObject = *objectsArrayPtr;
		}
		
//	NSLog(@"%s.. [%d] areaScore=%d, locationScore=%d, totalScore=%d", __PRETTY_FUNCTION__, objectArrayPos, areaScore, locationScore, score);
		
		if (bestObjectScore == 0)
			break;
	}
	
	return bestObject;
}

/**
 * Find the most prominent object in the array of objects. We do this when we
 * have found motion but are not currently tracking a particular object and we
 * need to pick an object to start tracked.
 *
 */
- (TrackedObject)findMostProminentObject:(TrackedObject *)objectsArray objectCount:(int)objectCount
{
	NSLog(@"%s.. looking at %d objects", __PRETTY_FUNCTION__, objectCount);
	
	int objectArrayPos = 0;
	TrackedObject *objectsArrayPtr = objectsArray;
	TrackedObject bestObject;
	
	memset(&bestObject, 0, sizeof(TrackedObject));
	
	for (; objectArrayPos < objectCount; ++objectArrayPos, ++objectsArrayPtr) {
		if (objectsArrayPtr->area > bestObject.area)
			bestObject = *objectsArrayPtr;
	}
	
	return bestObject;
}

/**
 * Calculates the center and area for each of the objects in the array 
 * objectsArray. Stops after the first size elements.
 *
 */
- (void)calculateCentersAndAreasForObjects:(TrackedObject *)objectsArray arraySize:(int)size
{
	int objectArrayLoc = 0;
	TrackedObject * object = objectsArray;
	
	for (; objectArrayLoc < size; ++objectArrayLoc, ++object) {
		[self calculateCenterAndAreaForObject:object];
	}
}

/**
 * Calculates the area of the object and its center point.
 *
 */
- (void)calculateCenterAndAreaForObject:(TrackedObject *)object
{
	int width = object->low - object->high;
	int height = object->right - object->left;
	
	object->area = 0;
	object->center.x = 0.0;
	object->center.y = 0.0;
	
	if (width <= 0 && height <= 0)
		return;
	
	object->area = width * height;
	object->center.x = object->right - (width / 2);
	object->center.y = object->low - (height / 2);
}

/**
 * Find the pixels that have changed beyond the allowable threshhold between
 * oldFrame and newFrame. These changes are noted in mMarks.
 *
 */
- (void)compareOldFrame:(unsigned int *)oldFrame withNewFrame:(unsigned int *)newFrame
{
//NSLog(@"%s..", __PRETTY_FUNCTION__);
	
	int row=0, col=0;
	
	unsigned int * oldFramePtr = oldFrame;
	unsigned int * newFramePtr = newFrame;
	
	CreepyColor difColor;
	CreepyColor oldColor;
	CreepyColor newColor;
	
	NSPoint foundPoint;
	
	// Look at each pixel in the frame, starting at the top left and ending at
	// the bottom right; moving left-to-right and top-to-bottom. if the pixel has
	// changed a lot, then try to figure out where that pixel moved to and mark
	// that location in the mMarks frame.
	//
	for (; row < IMAGE_ROWS; ++row) {
		for (col=0; col < IMAGE_COLS; ++col, ++oldFramePtr, ++newFramePtr) {
//		oldColor = *(CreepyColor *)oldFramePtr;
//		newColor = *(CreepyColor *)newFramePtr;
			oldColor = ((CreepyColor *)oldFramePtr)[0];
			newColor = ((CreepyColor *)newFramePtr)[0];
			
//		if (oldColor.r != 0 || newColor.r != 0)
//			NSLog(@"%s.. oldColor=0x%lx, newColor=0x%lx", __PRETTY_FUNCTION__, oldColor, newColor);
			
			// look at each color component of the pixel and add up the total change
			// for the pixel.
			//
			difColor.r = abs(oldColor.r - newColor.r);
			difColor.g = abs(oldColor.g - newColor.g);
			difColor.b = abs(oldColor.b - newColor.b);
			
//		if (difColor.r > 0 || difColor.g > 0 || difColor.b > 0)
//			NSLog(@"%s.. r=%d, g=%d, b=%d", __PRETTY_FUNCTION__, (int)difColor.r, (int)difColor.g, (int)difColor.b);
			
			// compare the sum of the color change to our color change threshold and
			// if the color change is insufficient, then ignore this pixel and move
			// on to the next.
			//
			if (difColor.r + difColor.g + difColor.b < COLOR_CHANGE_THRESHOLD)
				continue;
			
			// find where the pixel in oldFrame appears to have moved to in newFrame
			foundPoint = [self findPixelLike:oldColor nearRow:row andCol:col inFrame:newFrame];
			
//		NSLog(@"%s.. foundPoint.x=%d, y=%d", __PRETTY_FUNCTION__, (int)foundPoint.x, (int)foundPoint.y);
			
			// if we weren't able to find a suitably close pixel within our search 
			// area, just declare it a lost cause and move on to the next pixel.
			//
			if (foundPoint.x == -1.0 || foundPoint.y == -1.0)
				continue;
			
//		NSLog(@"%s.. I think that %d,%d moved to %d,%d", __PRETTY_FUNCTION__, col, row, (int)foundPoint.x, (int)foundPoint.y);
			
			// add destination in newFrame to mMarks
			mMarks[(int)(IMAGE_COLS*foundPoint.y)+(int)foundPoint.x].mark = 1;
		}
	}
}

/**
 * Always scan downward, including the current row (to the right).
 *
 */
- (NSPoint)findPixelLike:(CreepyColor)aColor nearRow:(int)aRow andCol:(int)aCol inFrame:(unsigned int *)aFrame
{
	int top, bot, left, right;
	int bestDistance, bestColorDif, bestMatch;
	NSPoint foundPoint = {-1.0, -1.0};
	
	bestDistance = 0xFFFF;
	bestColorDif = 0xFFFF;
	bestMatch    = 0xFFFF;
	
	// we want to scan SCAN_RANGE pixels UP and to the LEFT and RIGHT of our base
	// pixel location. we need to adjust that range if it goes off any edge of
	// the image.
	//
	// with the top row (top) being set to the row of the base pixel location, we
	// will end up scanning back SCAN_RANGE pixels on that row, but that 
	// shouldn't cause any catastrophic problems.
	//
	{
		top   = aRow;
		bot   = aRow + SCAN_RANGE;
		left  = aCol - SCAN_RANGE;
		right = aCol + SCAN_RANGE;
		
		if (top < 0) top = 0;
		if (bot >= IMAGE_ROWS) bot = IMAGE_ROWS - 1;
		if (left < 0) left = 0;
		if (right >= IMAGE_COLS) right = IMAGE_COLS - 1;
	}
	
	// scan in the top-to-bottom and left-to-right direction, and find the 
	// closest-most-similar (in terms of color) pixel.
	//
	{
		int scanRow, scanCol;
		
		for (scanRow=top; scanRow <= bot; ++scanRow) {
			for (scanCol=left; scanCol <= right; ++scanCol) {
				if (scanRow == aRow && scanCol == aCol)
					continue;
				
//			CreepyColor pixel = *(CreepyColor *)aFrame[(IMAGE_COLS*scanRow)+scanCol];
				CreepyColor pixel = ((CreepyColor *)aFrame)[(IMAGE_COLS*scanRow)+scanCol];
				
				// find distance from source pixel
				int distance = abs(scanRow - aRow) + abs(scanCol - aCol);
				
				// find similarity to target color
				int colorDif = abs((aColor.r - pixel.r) + 
													 (aColor.g - pixel.g) +
													 (aColor.b - pixel.b));
				
				// calculate the total similarity of the target pixel location and 
				// color, weighting the distance and color appropriately, to the pixel
				// that we're looking at right now.
				//
				int totalDif = (distance * DISTANCE_WEIGHT) + (colorDif * COLOR_WEIGHT);
				
//			NSLog(@"%s.. %d,%d = %lx,%lx,%lx; colorDif=%d, distance=%d, totalDif=%d", __PRETTY_FUNCTION__, 
//						scanCol, scanRow, (int)pixel.r, (int)pixel.g, (int)pixel.b, colorDif, distance, totalDif);
				
				// if this total similarity isn't any better than our current best, 
				// just move on to the next pixel.
				//
				if (totalDif >= bestMatch)
					continue;
				
				// since this pixel is all-around a better choice than our current 
				// best, replace all of these values accordingly.
				//
				bestMatch		 = totalDif;
				bestDistance = distance;
				bestColorDif = colorDif;
				foundPoint.x = scanCol;
				foundPoint.y = scanRow;
			}
		}
	}
	
	return foundPoint;
}

/**
 * Looks at all of the marked pixels in the given frame and groups them into
 * distinct objects based on the individual marks relative positions to each
 * other.
 *
 */
- (void)scanMarksForFrame:(CreepyPixel *)frame
{
	int row, col, objid;
	
	mTrackedObjectCount = 0;
	memset(&mTrackedObjects, 0, sizeof(TrackedObject) * MAX_TRACKED_OBJECTS);
	memset(&mMarkedPixelArray, 0, sizeof(NSPoint) * MARKED_PIXEL_ARRAY_SIZE);
	
	// iterate left-to-right and top-to-bottom
	//
	// if the pixel is marked
	//   if the pixel is not assigned to an object
	//     find any near-by marked pixel that belongs to an object and assign this one to the same object
	//     otherwise assign it to a new object
	//     assign all other near-by marked pixels to the same object
	//     update object top-bottom-left-right with new max locations as appropriate
	//
	// find centers of each object
	// compare center points
	//
	// if multiple objids are in the area, note that the objects might need to be combined.
	
	CreepyPixel * pixelPtr = mMarks;
	
	// for each pixel in the frame...
	for (row=0; row < IMAGE_ROWS; ++row) {
		for (col=0; col < IMAGE_COLS; ++col, ++pixelPtr) {
			// proceed to next loop if this pixel is not marked as having changed 
			// since the previous frame.
			if (!pixelPtr->mark)
				continue;
			
			// proceed to next loop if this pixel is already assigned to an object
			if (pixelPtr->object)
				continue;
			
			// get all near-by marked pixels and set objid to the object id of the 
			// closest of those near-by marked pixels.
			int nearByMarkedPixelCount = [self findNearByMarkedPixels:col vert:row objId:&objid array:mMarkedPixelArray arraySize:MARKED_PIXEL_ARRAY_SIZE];
			
			// NSLog(@"%s.. found %02d marked pixels near col=%03d, row=%03d, obj=%d", __PRETTY_FUNCTION__, nearByMarkedPixelCount, col, row, objid);
			
			// if there are no near-by marked pixels that are assigned to an object, 
			// create a new object.
			if (objid == 0) {
				if (mTrackedObjectCount >= MAX_TRACKED_OBJECTS) {
					NSLog(@"%s.. too many objects to track. sorry. max=%d", __PRETTY_FUNCTION__, MAX_TRACKED_OBJECTS);
					return;
				}
				
				TrackedObject * object = &mTrackedObjects[mTrackedObjectCount++];
//			TrackedObject * object = mTrackedObjects + mTrackedObjectCount++;
				objid = mTrackedObjectCount;
				
				object->high = row;
				object->low = row;
				object->left = col;
				object->right = col;
			}
			
			// set this current pixel to the object id
			pixelPtr->object = objid;
			
			// set all other marked-and-non-assigned pixels to this object id
			[self assignPixelsToObjectId:objid array:mMarkedPixelArray arraySize:nearByMarkedPixelCount frame:frame];
		}
	}
	
}

/**
 * Assuming a scan direction that moves left-to-right and top-to-bottom, look
 * backwards (ie, up) for any pixels near the pixel at row-x-col that are 
 * marked.
 *
 * Each such pixel should be added to array pointArray. The objid pointer 
 * should be set to the object id of the closest such pixel that has already
 * been assigned an object id.
 *
 */
- (int)findNearByMarkedPixels:(int)col vert:(int)row objId:(int *)objid array:(NSPoint *)pointArray arraySize:(int)arraySize
{
	int top, bot, left, right, closestObjId=0, pointArrayPos=0;
	
	// we want to scan SCAN_RANGE pixels UP and to the LEFT and RIGHT of our base
	// pixel location. we need to adjust that range if it goes off any edge of
	// the image.
	//
	// with the bottom row (bot) being set to the row of the base pixel location,
	// we will end up scanning ahead SCAN_RANGE pixels on that row, but that 
	// shouldn't cause any catastrophic problems.
	//
	{
		top   = row - SCAN_RANGE;
		bot   = row;
		left  = col - SCAN_RANGE;
		right = col + SCAN_RANGE;
		
		if (top < 0) top = 0;
		if (left < 0) left = 0;
		if (right >= IMAGE_COLS) right = IMAGE_COLS - 1;
	}
	
	// scan in the top-to-bottom and left-to-right direction, and add each marked
	// pixel to the pointArray array.
	//
	// if the pixel is already assigned an object id, then record that object id
	// in closestObjId as being the new and closest object id to our base pixel
	// location.
	//
	{
		int scanRow, scanCol;
		
		for (scanRow=top; scanRow <= bot; ++scanRow) {
			for (scanCol=left; scanCol <= right; ++scanCol) {
				int offset = (IMAGE_COLS * scanRow) + scanCol;
//			CreepyPixel *pixel = &mMarks[(IMAGE_COLS*scanRow)+scanCol];
//			CreepyPixel *pixel = &(CreepyPixel)mMarks[(IMAGE_COLS*scanRow)+scanCol];
//			CreepyPixel *pixel = (CreepyPixel *)&(CreepyPixel)(mMarks + ((IMAGE_COLS * scanRow) + scanCol));
				
//			if (!pixel->mark)
				if (!mMarks[offset].mark)
					continue;
				
				if (pointArrayPos < arraySize) {
					pointArray[pointArrayPos].x = scanCol;
					pointArray[pointArrayPos].y = scanRow;
					pointArrayPos++;
				}
				else
					NSLog(@"%s.. we've exceeded the capacity of pointArray, %d", __PRETTY_FUNCTION__, arraySize);
				
//			if (pixel->object)
				if (mMarks[offset].object)
//				closestObjId = pixel->object;
					closestObjId = mMarks[offset].object;
			}
		}
	}
	
	*objid = closestObjId;
	
	return pointArrayPos;
}

/**
 * For each NSPoint in pixelArray, get the corresponding pixel value frome our
 * target frame and if that pixel is not already assigned to an object, assign
 * it to objid.
 *
 */
- (void)assignPixelsToObjectId:(int)objid array:(NSPoint *)pixelArray arraySize:(int)arraySize frame:(CreepyPixel *)frame
{
	unsigned int pixelArrayPos = 0;
	CreepyPixel * pixel;
	NSPoint * pixelPtr = pixelArray;
	TrackedObject * object;
	
	// for point in the pixelArray, ...
	for (; pixelArrayPos < arraySize; ++pixelArrayPos, ++pixelPtr) {
		pixel = &frame[(IMAGE_COLS*(int)pixelPtr->y)+(int)pixelPtr->x];
		
		// if the pixel has already been assigned to an object, move on to the next
		// pixel.
		//
		// TODO
		//
		// we should look at noting associations between objects if they come very
		// close to each other; and this is the place for that.
		//
		if (pixel->object)
			continue;
		
		// assign this pixel its object id and get the object reference for that 
		// object id (which is one-indexed because I'm lazy).
		//
		pixel->object = objid;
		object = &mTrackedObjects[objid-1];
		
		// if this pixel's location exceeds the current bounding box for the object
		// to which it is now assigned, expand that bounding just enough to fit it.
		//
		if (pixelPtr->x < object->left)
			object->left = pixelPtr->x;
		else if (pixelPtr->x > object->right)
			object->right = pixelPtr->x;
		
		if (pixelPtr->y < object->high)
			object->high = pixelPtr->y;
		else if (pixelPtr->y > object->low)
			object->low = pixelPtr->y;
		
		// NSLog(@"%s.. [%02d of %02d] setting pixel at col=%d, row=%d to objid=%d", __PRETTY_FUNCTION__, pixelArrayPos, arraySize, (int)pixelPtr->x, (int)pixelPtr->y, objid);
	}
}

/**
 * Copy the new frame data into mFrameNew and compare this new frame with the 
 * previous frame. Then copy the new frame data over the old frame data.
 *
 */
- (NSPoint)addFrameData:(const unsigned char *)data width:(int)width height:(int)height bitsPerPixel:(int)bits
{
//NSLog(@"%s..", __PRETTY_FUNCTION__);
	
	if (IMAGE_COLS != width || IMAGE_ROWS != height || bits != 32) {
		NSLog(@"%s.. unsupported frame size, rows=%d, cols=%d, bits=%d", __PRETTY_FUNCTION__, height, width, bits);
		return (NSPoint){-1.0, -1.0};
	}
	
	memcpy(mFrameNew, data, (IMAGE_COLS * IMAGE_ROWS * 4));
	memset(mMarks, 0, sizeof(CreepyPixel) * IMAGE_COLS * IMAGE_ROWS);
	
	if (mFrameOld[0] != 0)
		[self compareOldFrame:mFrameOld withNewFrame:mFrameNew];
	
	memcpy(mFrameOld, mFrameNew, (IMAGE_COLS * IMAGE_ROWS * 4));
	[self scanMarksForFrame:mMarks];
	[self calculateCentersAndAreasForObjects:mTrackedObjects arraySize:mTrackedObjectCount];
	
	/*
	{
		NSLog(@"%s.. found %d distinct objects.", __PRETTY_FUNCTION__, mTrackedObjectCount);
		
		int objCnt = 0;
		
		for (; objCnt < mTrackedObjectCount; ++objCnt) {
			TrackedObject * object = &mTrackedObjects[objCnt];
			NSLog(@"%s..   [%02d of %02d] high=%d, low=%d, left=%d, right=%d, area=%d, center.x=%d, center.y=%d", __PRETTY_FUNCTION__, 
						(objCnt+1), mTrackedObjectCount, object->high, object->low, object->left, object->right,
						object->area, (int)object->center.x, (int)object->center.y);
		}
	}
	*/
	
	// find where we think our currently tracked object has moved to,
	// if it has moved at all, or assign one of the objects to be the
	// currently tracked object if we don't already have one assigned.
	//
	if (mTrackedObjectCount == 0)
		NSLog(@"%s.. we couldn't find any movement. sorry.", __PRETTY_FUNCTION__);
	else if (mCurTrackedObject.center.x == 0.0 || mCurTrackedObject.center.y == 0.0)
		mCurTrackedObject = [self findMostProminentObject:(TrackedObject *)mTrackedObjects objectCount:mTrackedObjectCount];
	else
		mCurTrackedObject = [self findObjectLike:mCurTrackedObject fromObjects:mTrackedObjects objectCount:mTrackedObjectCount];
	
	return mCurTrackedObject.center;
}

- (NSPoint)__addFrameData:(const unsigned char *)data width:(int)width height:(int)height bitsPerPixel:(int)bits
{
//NSPoint point;
	
	if (mLastPoint.x >= width-1)
		mHorzDir = 0;
	else if (mLastPoint.x == 0)
		mHorzDir = 1;
	
	if (mLastPoint.y >= height-1)
		mVertDir = 0;
	else if (mLastPoint.y == 0)
		mVertDir = 1;
	
	int dimensionToMove = rand() % 2;
	
	if (dimensionToMove == 0)
		mLastPoint.x += mHorzDir == 0 ? -10 : 10;
	else if (dimensionToMove == 1)
		mLastPoint.y += mVertDir == 0 ? -10 : 10;
	
	if (mLastPoint.x >= width)
		mLastPoint.x = width-1;
	else if (mLastPoint.x < 0)
		mLastPoint.x = 0;
	
	if (mLastPoint.y >= height)
		mLastPoint.y = height-1;
	else if (mLastPoint.y < 0)
		mLastPoint.y = 0;
	
	/*
	const unsigned int *image = (const unsigned int *)data;
	
	int col = 99;
	int row = 1;
	
	image[(width*row) + col];
	*/
	
	return mLastPoint;
}

@end
